#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "resource.h"
#include "strings.h"

/* HTTP-Header parsen und RequestInfo erneuern */
int parseheader(char * buffer, struct ReqInfo * reqinfo) {
    static int header = 1;
    char      *temp;
    char      *endptr;
    int        len;

    if (header == 1) {
	/* Wenn header 1 ist, ist das die Zeile, welche die Request enthlt */

		/* HTTP-Methode erkennen (GET oder HEAD) */
		if (!strncmp(buffer, "GET ", 4)) {
			reqinfo->method = GET;
			buffer += 4;
		} else if (!strncmp(buffer, "HEAD ", 5)) {
			reqinfo->method = HEAD;
			buffer += 5;
		} else {
			reqinfo->method = UNSUPPORTED;
			reqinfo->status = 501;
			return -1;
		}

		/* Solange weiter nach Rechts bewegen bis die erste Ressource erreicht ist */
		while (*buffer && isspace(*buffer))
		buffer++;
		
		/* Lnge des Ressourcen-Strings ermitteln */
		endptr = strchr(buffer, ' ');
		if (endptr == NULL) {
			len = strlen(buffer);
		} else {
			len = endptr - buffer;
		}
		
		if (len == 0) {
			reqinfo->status = 400;
			return -1;
		}
		
		/* String in der RequestInfo Struktur speichern */
		reqinfo->resource = calloc(len + 1, sizeof(char));
		strncpy(reqinfo->resource, buffer, len);

	
		/* Der Server unterscheided zwischen vollstndigen und simplen Anfragen */

		if ( strstr(buffer, "HTTP/") )
			reqinfo->type = FULL;
		else
			reqinfo->type = SIMPLE;

		header = 0;
		return 0;
    }


    /* Wenn wir hier ankommen, handelt es sich um eine komplette HTTP-Anfrage, die jetzt geparset werden muss */

	/* Anfrage berprfen und alle Zeichen zu Grobuchstaben umwandeln */
    endptr = strchr(buffer, ':');
    if (endptr == NULL) {
		reqinfo->status = 400;
		return -1;
    }

    temp = calloc((endptr - buffer) + 1, sizeof(char));
    strncpy(temp, buffer, (endptr - buffer));
    strupper(temp);

    buffer = endptr + 1;
    while ( *buffer && isspace(*buffer) )
	++buffer;
    if ( *buffer == '\0' )
     	return 0;

    if ( !strcmp(temp, "USER-AGENT") ) {
	    reqinfo->useragent = malloc( strlen(buffer) + 1 );
	    strcpy(reqinfo->useragent, buffer);
    }
    else if ( !strcmp(temp, "REFERER") ) {
	    reqinfo->referer = malloc( strlen(buffer) + 1 );
	    strcpy(reqinfo->referer, buffer);
    }

    free(temp);
    return 0;
}


/* Liest die Anfrage des Clients, hier ist ein Timeout ntig da wir in seltenen Fllen sonst ewig warten knnten */
int getrequest(int conn, struct ReqInfo * reqinfo) {

    char   buffer[REQSIZE] = {0};
    int    rval;
    fd_set fds;
    struct timeval tv;

    /* Timeout: 10s */

    tv.tv_sec  = 10;
    tv.tv_usec = 0;

    do {
		FD_ZERO(&fds);
		FD_SET (conn, &fds);

		rval = select(conn + 1, &fds, NULL, NULL, &tv);

		if (rval < 0) {
			quit_error(E_SOCKET_REQUEST);
		}
		else if (rval == 0) {
			return -1;
		}
		else {
			socket_readline(conn, buffer, REQSIZE - 1);
			trim(buffer);

			if (buffer[0] == '\0')
			break;

			if (parseheader(buffer, reqinfo))
			break;
		}
    } while (reqinfo->type != SIMPLE);

    return 0;
}


/* RequestInfo Struktur initialisieren */
void reginfoinit(struct ReqInfo * reqinfo) {
    reqinfo->useragent = NULL;
    reqinfo->referer   = NULL;
    reqinfo->resource  = NULL;
    reqinfo->method    = UNSUPPORTED;
    reqinfo->status    = 200;
}

	
/* RequestInfo Struktur freigeben */
void reginfofree(struct ReqInfo * reqinfo) {
    if ( reqinfo->useragent )
	free(reqinfo->useragent);
    if ( reqinfo->referer )
	free(reqinfo->referer);
    if ( reqinfo->resource )
	free(reqinfo->resource);
}


/* HTTP-Header fr Antwort ausgeben */
int outputheaders(int conn, struct ReqInfo * reqinfo) {
    char buffer[100];
	
    sprintf(buffer, "HTTP/1.0 %d OK\r\n", reqinfo->status);
    socket_writeline(conn, buffer, strlen(buffer));
    socket_writeline(conn, "Server: uhttps 0.1.0\r\n", 22);
    /* TODO: MIME Untersttzung
	 * Browser sollen im Moment selbst die Erkennung vornehmen.
	 * socket_writeline(conn, "Content-Type: text/html\r\n", 25); */
    socket_writeline(conn, "\r\n", 2);
	
    return 0;
}


/* Verarbeitet eine HTTP-Request */
int serverrequest(int conn) {
    struct ReqInfo  reqinfo;
    int             resource = 0;

    reginfoinit(&reqinfo);
    
    /* Request lesen */
    if (getrequest(conn, &reqinfo) < 0)
	return -1;
    
    /* Request verarbeiten und ggf. HTTP-Status neu setzen */
    if (reqinfo.status == 200) {
		if ((resource = checkfile(&reqinfo)) < 0) {
			if (errno == EACCES)
				reqinfo.status = 401;
			else
				reqinfo.status = 404;
		}
	}

    /* Header ausgeben wenn Request komplett */
    if (reqinfo.type == FULL)
	outputheaders(conn, &reqinfo);

	/* Antwort inklusive aufgerufener Seite abschicken */
    if (reqinfo.status == 200) {
	if (returnfile(conn, resource, &reqinfo))
	    quit_error(E_RESOURCE_RETURN);
    }
    else
	returnerror(conn, &reqinfo);

    if (resource > 0)
	if (close(resource) < 0)
	    quit_error(E_RESOURCE_CLOSE);
	
    reginfofree(&reqinfo);

    return 0;
}
