#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "strings.h"

/* Leerzeichen hinter einem String entfernen */
int trim(char * buffer) {
    int n = strlen(buffer) - 1;

    while (!isalnum(buffer[n]) && n >= 0)
		buffer[n--] = '\0';

    return 0;
}


/* String in Grobuchstaben umformen */
int strupper(char * buffer) {
    while (*buffer) {
		*buffer = toupper(*buffer);
		++buffer;
    }
	
    return 0;
}


/* URL-Encode zurckbersetzen */
void translate_url(char * buffer) {
    char asciinum[3] = {0};
    int i = 0, c;
    
    while (buffer[i]) {
		if (buffer[i] == '+')
			buffer[i] = ' ';
		else if (buffer[i] == '%') {
			asciinum[0] = buffer[i+1];
			asciinum[1] = buffer[i+2];
			buffer[i] = strtol(asciinum, NULL, 16);
			c = i+1;
			do {
				buffer[c] = buffer[c+2];
			} while (buffer[2+(c++)]);
		}
		++i;
    }
}


/* Gibt einen Fehler aus und beendet das Programm */
void quit_error(char const * message) {
    fprintf(stderr, "Fehler: %s\n", message);
    exit(EXIT_FAILURE);
}

/* Gibt einen Hinweis aus */
void notice(char const * message) {
    fprintf(stderr, "%s\n", message);
}

/* Socket zeilenweise auslesen */
ssize_t socket_readline(int sockd, void *vptr, size_t maxlen) {
    ssize_t n, rc;
    char    c, *buffer;

    buffer = vptr;
	
    for (n = 1; n < maxlen; n++) {
		if ((rc = read(sockd, &c, 1)) == 1) {
			*buffer++ = c;
			if (c == '\n')
			break;
		} else if (rc == 0) {
			if (n == 1)
			return 0;
			else
			break;
		} else {
			if (errno == EINTR)
			continue;
			quit_error(E_SOCKET_READ);
		}
    }

    *buffer = 0;
    return n;
}


/* Zeilenweise in Socket schreiben */
ssize_t socket_writeline(int sockd, const void *vptr, size_t n) {
    size_t      nleft;
    ssize_t     nwritten;
    const char *buffer;

    buffer = vptr;
    nleft  = n;

    while (nleft > 0) {
		if ((nwritten = write(sockd, buffer, nleft)) <= 0 ) {
			if (errno == EINTR)
			nwritten = 0;
			else
			quit_error(E_SOCKET_WRITE);
		}
	nleft  = nleft - nwritten;
	buffer = buffer + nwritten;
    }

    return n;
}
